/*
   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.
*/


#include <windows.h>
#include <windowsx.h>
#include <objbase.h>
#include <commctrl.h>
#include <shlobj.h>

LPITEMIDLIST GetIDListFromPath(HWND hWnd, LPCSTR lpPath, ULONG dwAttribs)
{
	IShellFolder *psf = NULL;
	LPITEMIDLIST lpidl = NULL;
	HRESULT hres;
	hres = SHGetDesktopFolder(&psf);
	if (SUCCEEDED(hres)) {
		WCHAR lpDisplayName[_MAX_PATH];
		MultiByteToWideChar(CP_OEMCP, 0, lpPath, _MAX_PATH, lpDisplayName, _MAX_PATH);
		hres = psf->lpVtbl->ParseDisplayName(psf, hWnd, NULL, lpDisplayName, NULL, &lpidl, &dwAttribs);
		psf->lpVtbl->Release(psf);
	}
	return lpidl;
}


int CALLBACK BrowseCallbackProc(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
	switch (uMsg) {
	case BFFM_INITIALIZED:
		if (lpData != (LPARAM)NULL) {
			SendMessage(hwnd, BFFM_SETSELECTION, TRUE, lpData);
		}
		break;
	case BFFM_SELCHANGED:
		if (lParam != (LPARAM)NULL) {
			CHAR chFolder[_MAX_PATH];
			if (SHGetPathFromIDList ((LPITEMIDLIST)lParam, chFolder)) {
				SendMessage(hwnd, BFFM_SETSTATUSTEXT, (LPARAM)NULL, (LPARAM)chFolder);
			}
		}
		break;
	}
	return 0;
}


BOOL SelectDirectoryOpt(HWND hParent, LPCSTR lpCaption, LPCSTR lpRoot, LPSTR lpDirectory, UINT ulFlags)
{
	BROWSEINFO bI;
	LPITEMIDLIST pidlRetBrowse = NULL, pidlRoot = NULL;
	
	if (lpRoot != NULL) {
		pidlRoot = GetIDListFromPath(hParent, lpRoot, 0);
	}
	ZeroMemory(&bI,sizeof(bI));
	bI.hwndOwner = hParent;
	bI.pidlRoot = pidlRoot;
	bI.pszDisplayName = lpDirectory;
	bI.lpszTitle = lpCaption;
	bI.ulFlags = ulFlags | BIF_STATUSTEXT;
	bI.lpfn = BrowseCallbackProc;
	bI.lParam = (LPARAM)lpDirectory;
	pidlRetBrowse = SHBrowseForFolder(&bI);
	if (pidlRetBrowse != NULL) {
		SHGetPathFromIDList(pidlRetBrowse, lpDirectory);
		CoTaskMemFree(pidlRetBrowse);
	}
	if (lpRoot != NULL) {
		CoTaskMemFree(pidlRoot);
	}
	return pidlRetBrowse != NULL;
}


BOOL SelectDirectory(HWND hParent, LPCSTR lpCaption, LPCSTR lpRoot, LPSTR lpDirectory)
{
	return SelectDirectoryOpt(hParent, lpCaption, lpRoot, lpDirectory, BIF_RETURNONLYFSDIRS);
}
