/*
  This file is under public domain.
*/

/*
  references
  
  SendKeys Method
  http://msdn.microsoft.com/en-us/library/8c6yea83(VS.85).aspx
  Windows Script Host
  http://msdn.microsoft.com/en-us/library/9bbdkx3k(VS.85).aspx
  WshShell Object
  http://msdn.microsoft.com/en-us/library/aew9yb99(VS.85).aspx
  AutoHotkey - Free Mouse and Keyboard Macro Program with Hotkeys and AutoText
  http://www.autohotkey.com/
  Hotkeys (Mouse, Joystick and Keyboard Shortcuts)
  http://www.autohotkey.com/docs/Hotkeys.htm
  VkKeyScan Function ()
  http://msdn.microsoft.com/en-us/library/ms646329(VS.85).aspx
  GetKeyState Function (Windows)
  http://msdn.microsoft.com/en-us/library/ms646301(VS.85).aspx
  Keyboard Input
  http://msdn.microsoft.com/en-us/library/ms645530(VS.85).aspx
  Sorting it all Out : What the %$#!* is wrong with VkKeyScan[Ex]?
  http://blogs.msdn.com/michkap/archive/2006/03/26/560595.aspx
  
*/

#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdio.h>

#define ARRAYSIZE(a) (sizeof(a) / sizeof(*(a)))

static struct {
	CONST TCHAR * CONST name;
	CONST WORD vk;
	size_t length;
} g_keywords[] = {
	{_T("BACKSPACE"), VK_BACK},
	{_T("BS"), VK_BACK},
	{_T("BKSP"), VK_BACK},
	{_T("BREAK"), VK_CANCEL},
	{_T("CAPSLOCK"), VK_CAPITAL},
	{_T("DELETE"), VK_DELETE},
	{_T("DEL"), VK_DELETE},
	{_T("DOWN"), VK_DOWN},
	{_T("END"), VK_END},
	{_T("ENTER"), VK_RETURN},
	{_T("ESC"), VK_ESCAPE},
	{_T("HELP"), VK_HELP},
	{_T("HOME"), VK_HOME},
	{_T("INSERT"), VK_INSERT},
	{_T("INS"), VK_INSERT},
	{_T("LEFT"), VK_LEFT},
	{_T("NUMLOCK"), VK_NUMLOCK},
	{_T("PGDN"), VK_NEXT},
	{_T("PGUP"), VK_PRIOR},
	{_T("PRTSC"), VK_SNAPSHOT},
	{_T("RIGHT"), VK_RIGHT},
	{_T("SCROLLLOCK"), VK_SCROLL},
	{_T("TAB"), VK_TAB},
	{_T("UP"), VK_UP},
	{_T("F1"), VK_F1},
	{_T("F2"), VK_F2},
	{_T("F3"), VK_F3},
	{_T("F4"), VK_F4},
	{_T("F5"), VK_F5},
	{_T("F6"), VK_F6},
	{_T("F7"), VK_F7},
	{_T("F8"), VK_F8},
	{_T("F9"), VK_F9},
	{_T("F10"), VK_F10},
	{_T("F11"), VK_F11},
	{_T("F12"), VK_F12},
	{_T("F13"), VK_F13},
	{_T("F14"), VK_F14},
	{_T("F15"), VK_F15},
	{_T("F16"), VK_F16},
	{_T("F17"), VK_F17},
	{_T("F18"), VK_F18},
	{_T("F19"), VK_F19},
	{_T("F20"), VK_F20},
	{_T("F21"), VK_F21},
	{_T("F22"), VK_F22},
	{_T("F23"), VK_F23},
	{_T("F24"), VK_F24},
};


/*
   See the source and comment of AutoHotKey,
   about definition of IsKeyDown9xNT and IsKeyDown2kXP in keyboard_mouse.h.
 */
static int (*g_is_key_pushed)(int);

static int is_key_pushedW(int k)
{
	return GetKeyState(k) < 0;
}

static int is_key_pushedA(int k)
{
	return GetAsyncKeyState(k) < 0 || GetKeyState(k) < 0;
}

#ifdef ISOLATED_TEST
#include <limits.h>
#include "bitvector.h"
static unsigned char g_keyflag[256 / CHAR_BIT];
static int is_key_pushed_dummy(int k)
{
	if (k < 1 || k > 254) return -1;
	return bitvector_get(g_keyflag, k);
}
#endif /* ISOLATED_TEST */


static void (*g_sendkey)(WORD, WORD, DWORD, DWORD);

#if (_WIN32_WINNT >= 0x0403)
static void sendkeyW(WORD vkey, WORD scan, DWORD flags, DWORD info)
{
	INPUT input;
	input.type = INPUT_KEYBOARD;
	input.ki.wVk = vkey;
	input.ki.wScan = scan;
	input.ki.dwFlags = flags;
	input.ki.time = 0;
	input.ki.dwExtraInfo = info;
	SendInput(1, &input, sizeof(INPUT));
}
#endif /* (_WIN32_WINNT >= 0x0403) */


static void sendkeyA(WORD vkey, WORD scan, DWORD flags, DWORD info)
{
	keybd_event((BYTE) vkey, (BYTE) scan, flags, info);
}


#ifdef ISOLATED_TEST
static void sendkey_print(WORD vkey, WORD scan, DWORD flags, DWORD info)
{
	bitvector_set(g_keyflag, vkey, ! (flags & KEYEVENTF_KEYUP));
	_tprintf("%x %x %lx %lx\n", vkey, scan, flags, info);
}
#endif /* ISOLATED_TEST */


static const BYTE g_states[] = {VK_SHIFT, VK_CONTROL, VK_MENU};
enum { SHIFT = 1, CTRL = 2, ALT = 4};
enum { OP_NONE, OP_PUSH, OP_RELEASE};


static BYTE single_modifier(BYTE mod, int op)
{
	int i;
	if (mod == 0) return 0;
	i = mod >> 1;
	switch (op) {
	case OP_PUSH:
		if ((*g_is_key_pushed)(g_states[i])) return 0;
		(*g_sendkey)(g_states[i], 0, 0, 0);
		break;
	case OP_RELEASE:
		if ((*g_is_key_pushed)(g_states[i])) {
			(*g_sendkey)(g_states[i], 0, KEYEVENTF_KEYUP, 0);
		}
		break;
	}
	return mod;
}


static WORD modifier(WORD mod, int op)
{
	size_t i;
	WORD mask = 1u;
	if (mod == 0) return 0;
	for (i = 0; i < ARRAYSIZE(g_states); i ++) {
		if (mask & mod) {
			if ((*g_is_key_pushed)(g_states[i])) {
				if (op == OP_RELEASE) {
					(*g_sendkey)(g_states[i], 0, KEYEVENTF_KEYUP, 0);
				}
				else {
					mod &= ~mask;
				}
			}
			else if (op == OP_PUSH) {
				(*g_sendkey)(g_states[i], 0, 0, 0);
			}
		}
		mask <<= 1;
	}
	return mod;
}


static WORD key(int vk, WORD mod, int op)
{
	switch (op) {
	case OP_PUSH:
		if ((*g_is_key_pushed)(vk)) return 0;
		mod = modifier(mod, op);
		(*g_sendkey)(vk, 0, 0, 0);
		break;
	case OP_RELEASE:
		if (! (*g_is_key_pushed)(vk)) return 0;
		(*g_sendkey)(vk, 0, KEYEVENTF_KEYUP, 0);
		mod = modifier(mod, op);
		break;
	}
	return MAKEWORD(vk, mod);
}


static WORD word_key(WORD k, int op)
{
	return key(LOBYTE(k), HIBYTE(k), op);
}


static SHORT char_key(TCHAR ch, int op)
{
	SHORT k;
	BYTE vk, m;
	
	k = VkKeyScan(ch);
	vk = LOBYTE(k);
	m = HIBYTE(k);
	if (vk == (BYTE) -1 && m == (BYTE) -1) return -1;
	return key(vk, m, op);
}


static SHORT brace(LPTSTR *keys)
{
	LPTSTR c = *keys, p;
	size_t i, length;
	SHORT k = MAKEWORD(-1, -1);
	long n;
	BYTE vk, m;
	
	if (*c == _T('}')) {
		k = VkKeyScan(_T('}'));
		c = CharNext(c);
	}
	else {
		for (p = c; *p != _T(' ') && *p != _T('}'); p = CharNext(p)) {
			if (! *p) return -1;
		}
		length = p - c;
		if (length == 1) {
			k = VkKeyScan(*c);
		}
		else {
			for (i = 0; i < ARRAYSIZE(g_keywords); i ++) {
				if (g_keywords[i].length == length &&
				    _tcsncmp(c, g_keywords[i].name, length) == 0) {
					k = g_keywords[i].vk;
					break;
				}
			}
			if (i == ARRAYSIZE(g_keywords)) {
				n = _tcstol(c, &c, 0);
				if (c != p || n < 1 || n > 254)  return -1;
				k = (SHORT) n;
			}
		}
		c = p;
	}
	vk = LOBYTE(k);
	m = HIBYTE(k);
	if (vk != (BYTE) -1 && m != (BYTE) -1) {
		if (*c == _T('}')) {
			*keys = CharNext(c);
			return key(vk, m, OP_PUSH);
		}
		else if (*c == _T(' ')) {
			c = CharNext(c);
			if (*c) {
				n = _tcstol(c, &c, 10);
				if (*c == _T('}')) {
					for (; n > 0; n --) {
						k = word_key(k, OP_PUSH);
						k = word_key(k, OP_RELEASE);
					}
					*keys = CharNext(c);
					return 0;
				}
			}
		}
	}
	return -1;
}


LPTSTR sendkey_eval(LPTSTR keys)
{
	SHORT k = 0;
	WORD m = 0, w = 0;
	
	while (keys && *keys && k >= 0) {
		if (*keys == _T(')')) {
			keys = CharNext(keys);
			break;
		}
		if (*keys == _T('}')) {
			keys = NULL;
			break;
		}
		if (*keys == _T('(')) {
			keys = sendkey_eval(CharNext(keys));
			if (k > 0) {
				word_key(k, OP_RELEASE);
				k = 0;
			}
			single_modifier(m, OP_RELEASE);
			m = 0;
			continue;
		}
		if (k > 0) {
			word_key(k, OP_RELEASE);
			k = 0;
		}
		if (m > 0 && w > 0) {
			single_modifier(m, OP_RELEASE);
			m = 0;
		}
		w = 1;
		if (*keys == _T('{')) {
			keys = CharNext(keys);
			k = brace(&keys);
			continue;
		}
		if (*keys == _T('+') || *keys == _T('^') || *keys == _T('%')) {
			single_modifier(m, OP_RELEASE);
			switch (*keys) {
			case _T('+'):
				m = SHIFT;
				break;
			case _T('^'):
				m = CTRL;
				break;
			case _T('%'):
				m = ALT;
				break;
			}
			m = single_modifier(m, OP_PUSH);
			w = 0;
		}
		else if (*keys == _T('~')) {
			k = key(VK_RETURN, 0, OP_PUSH);
		}
		else {
			k = char_key(*keys, OP_PUSH);
		}
		keys = CharNext(keys);
	}
	if (k > 0) {
		word_key(k, OP_RELEASE);
	}
	if (m > 0) {
		single_modifier(m, OP_RELEASE);
	}
	return keys;
}


LPTSTR sendkey_raw(LPTSTR keys)
{
	WORD vkey = 0, scan = 0;
	DWORD flags = 0, info = 0;
	LPTSTR p;
	unsigned long n;
	int i;
	
	for (i = 0; *keys; i = (i + 1) & 3) {
		n = _tcstoul(keys, &p, 0);
		if (keys == p) break;
		keys = p;
		switch (i) {
		case 0:
			vkey = n;
			break;
		case 1:
			scan = n;
			break;
		case 2:
			flags = n;
			break;
		case 3:
			info = n;
			(*g_sendkey)(vkey, scan, flags, info);
			break;
		}
	}
	return keys;
}


void sendkey_init(void)
{
	size_t i;
	
	if (! g_sendkey) {
		OSVERSIONINFO ovi;
		ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		g_sendkey = sendkeyA;
		g_is_key_pushed = is_key_pushedA;
		if (GetVersionEx(&ovi) && ovi.dwPlatformId == VER_PLATFORM_WIN32_NT) {
#if (_WIN32_WINNT >= 0x0403)
			g_sendkey = sendkeyW;
#endif /* (_WIN32_WINNT >= 0x0403) */
			if (ovi.dwMajorVersion >= 5) {
				g_is_key_pushed = is_key_pushedW;
			}
		}
		for (i = 0; i < ARRAYSIZE(g_keywords); i ++) {
			g_keywords[i].length = _tcslen(g_keywords[i].name);
		}
	}
}


#ifdef ISOLATED_TEST
int main(int argc, char *argv[])
{
	int c1, c2, r;
	int i;
	char *pat[] = {
		"^n", "^(n)", "^(^n)", "^{n}", "^{(n)}", "^({n})", "+^%n", "^{n 5}",
		"^{no}", "^{F1", "^{F", "^{F ", "^{F 3", "^{F }", "n^{bc }a", 
		"{", "}", "(", ")", "}notreach",
		 "^+%", "n(a^bc)n", "^{F1}", "^{BSP}", "^{BS P}",
		"n(c^na", "^{F2 3}", "+{} 3}",
		"%~", "^n(mn^n){)}{}}", "+n(mn^n{)}{}})",
		"{} 2}", "{{ 2}", "{}", "{{{", "{{2", "{}2", 
		NULL,
	};
	c1 = c2 = 0;
	g_sendkey = sendkey_print;
	g_is_key_pushed = is_key_pushed_dummy;
	printf("%d\n", ARRAYSIZE(g_keywords));
	for (i = 0; i < ARRAYSIZE(g_keywords); i ++) {
		g_keywords[i].length = _tcslen(g_keywords[i].name);
	}
	/*
	for (i = 0; i < ARRAYSIZE(g_keywords); i ++) {
		char buffer[10];
		char *p = buffer;
		int k;
		sprintf(buffer, "%s}", g_keywords[i].name);
		k = brace(&p);
		r = (k == g_keywords[i].vk);
		if (! r) c1 ++;
		printf("%u %s -> %x %x %d\n", g_keywords[i].length, buffer, k, g_keywords[i].vk, r);
	}
	for (i = 1; i < 254; i ++) {
		char buffer[10];
		char *p = buffer;
		int k;
		sprintf(buffer, "0x%x}", i);
		k = brace(&p);
		r = (i == k);
		if (! r) c2 ++;
		printf("%x -> %x %d\n", i, k, r);
	}
	*/
	printf("%d %d\n", c1, c2);
	for (i = 0; pat[i]; i ++) {
		printf("pat %d %s\n", i, pat[i]);
		sendkey_eval(pat[i]);
	}
	return 0;
}
#endif /* ISOLATED_TEST */
