/*
  This file is under public domain.
*/
#include <windows.h>
#include <tchar.h>

#include <stdlib.h>
#include <string.h>

#include "utils.h"


LPTSTR normalized_path(LPCTSTR lpszPath)
{
	static const LPTSTR prefix = TEXT("\\\\?\\");
	static const int length = lstrlen(prefix);
	LPTSTR buffer, front, p, pp;
	
	if (lpszPath == NULL) return NULL;
	/* duplicate path string */
	buffer = _tcsdup(lpszPath);
	if (buffer == NULL) return NULL;
	/* skip wide prefix */
	front = buffer;
	if (_tcsncmp(front, prefix, length) == 0) front += length;
	/* replace slash to backslash and remove continuous backslash */
	for (p = front; *p != 0; ) {
		if (*p == TEXT('\\') || *p == TEXT('/')) {
			if (*p == TEXT('/')) *p = TEXT('\\');
			p = CharNext(p);
			if (*p == TEXT('\\') || *p == TEXT('/')) {
				for (pp = p; *pp == TEXT('\\') || *pp == TEXT('/'); pp = CharNext(pp))
					;
				lstrcpy(p, pp);
			}
		}
		else {
			p = CharNext(p);
		}
	}
	/* remove tail backslash */
	p = CharPrev(front, p);
	while (p > front && *p == TEXT('\\')) {
		*p = 0;
		p = CharPrev(front, p);
	}
	return buffer;
}


BOOL create_directory(LPTSTR lpszPath, LPSECURITY_ATTRIBUTES lpsa)
{
	LPTSTR c;
	
	if (lpszPath == NULL || *lpszPath == 0)
		return FALSE;
	if (CreateDirectory(lpszPath, lpsa))
		return TRUE;
	if (GetLastError() != ERROR_PATH_NOT_FOUND)
		return FALSE;
	
	/* search last backslash */
	c = lpszPath + lstrlen(lpszPath) - 1;
	for (; *c != TEXT('\\'); c = CharPrev(lpszPath, c))
		if (c == lpszPath)
			return FALSE;
	
	*c = TEXT('\0');
	if (! create_directory(lpszPath, lpsa))
		return FALSE;
	*c = TEXT('\\');
	return CreateDirectory(lpszPath, lpsa);
}


void slash2backslash(LPTSTR p)
{
	for (; *p != 0; p = CharNext(p))
		if (*p == TEXT('/')) *p = TEXT('\\');
}


TCHAR *next_line(TCHAR **lines, TCHAR **buffer, size_t *length)
{
	TCHAR *p;
	size_t i;
	
	p = _tcspbrk(*lines, _T("\r\n"));
	if (p == NULL) {
		p = *lines;
		*lines = NULL;
		return p;
	}
	i = p - *lines + 1;
	if (*length < i) {
		free(*buffer);
		TCHAR *b = (TCHAR *) malloc(i * sizeof(TCHAR));
		if (b == NULL) return NULL;
		*buffer = b;
		*length = i;
	}
	memcpy(*buffer, *lines, i * sizeof(TCHAR));
	(*buffer)[i - 1] = _T('\0');
	*lines = p + ((p[0] == _T('\r') && p[1] == _T('\n')) ? 2 : 1);
	return *buffer;
}


TCHAR *ul_to_as(ULONG num)
{
	TCHAR *buffer;
	ULONG n;
	int i;
	
	if (num > 0) {
		for (n = num, i = 0; n > 0; n /= 10, i ++)
			;
	} else {
		i = 1;
	}
	buffer = (TCHAR *) malloc((i + 1) * sizeof(TCHAR));
	if (buffer == NULL) return NULL;
	buffer[i] = _T('\0');
	do {
		buffer[-- i] = _T('0') + num % 10;
	} while ((num /= 10) > 0);
	return buffer;
}
