/*
  This file is under public domain.
*/

/*
  references
  
  document.evaluate - MDC
  https://developer.mozilla.org/en/DOM/document.evaluate
  
  seamonkey source
    extensions/transformiix/source/xpath/nsXPathResult.h
    extensions/transformiix/source/xpath/nsXPathResult.cpp
  
  Gecko SDK
    include\dom\nsIDOMXPathEvaluator.h
    include\dom\nsIDOMXPathResult.h
    include\dom\nsIDOMNSHTMLElement.h
  
 */

/*
  Debug Helper
  return km.pluginmsgex("utils", "xpathget", "1 + 1", km.TYPE_STRING)
  return km.pluginmsgex("utils", "xpathget", "1 div 2", km.TYPE_STRING)
  return km.pluginmsgex("utils", "xpathget", "//p", km.TYPE_STRING)
  return km.pluginmsgex("utils", "xpathget", "count(//p)", km.TYPE_STRING)
  return km.pluginmsgex("utils", "xpathget", "//p[position()=2]", km.TYPE_STRING)
  return km.pluginmsg("utils", "xpathset", "//p", "test<br />test")
  return km.pluginmsg("utils", "xpathset", "//a/@href", "#top")
 */

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>

#include <nspr.h>
#include <nsIWebBrowser.h>
#include <nsIDOMWindow.h>
#include <nsIDOMDocument.h>
#include <nsIDOMNode.h>
#include <nsIDOMNSHTMLElement.h>
#include <nsIDOMXPathEvaluator.h>
#include <nsIDOMXPathResult.h>
#include <nsEmbedString.h>

#include "miniCOMPtr.h"
#include "miniMBString.h"
#include "miniWString.h"
#include "utils.h"

extern kmeleonPlugin kPlugin;

#define MAX_DELIMITER_LENGTH 11
static PRUnichar g_delimiter[MAX_DELIMITER_LENGTH];


static BOOL xpath_eval(HWND hwnd, char *expr,
                       miniCOMPtr<nsIDOMXPathResult>& result)
{
	nsresult rv;
	
	miniCOMPtr<nsIWebBrowser> browser;
	if (! kPlugin.kFuncs->GetMozillaWebBrowser(hwnd, &browser)) {
		return FALSE;
	}
	
	miniCOMPtr<nsIDOMWindow> domwin;
	rv = browser->GetContentDOMWindow(&domwin);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(domwin, FALSE);
	
	miniCOMPtr<nsIDOMDocument> doc;
	rv = domwin->GetDocument(&doc);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(doc, FALSE);
	
	miniCOMPtr<nsIDOMXPathEvaluator> evaluator;
	rv = doc.QueryInterface(evaluator);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(evaluator, FALSE);
	
	nsEmbedString expression;
	expression = miniWString(expr);
	rv = evaluator->Evaluate(expression, static_cast<nsIDOMDocument *>(doc), NULL, nsIDOMXPathResult::ANY_TYPE, NULL, reinterpret_cast<nsISupports **>(&result));
	if (NS_SUCCEEDED(rv)) return TRUE;
	MessageBox(hwnd, translate(LABEL_XPATH_ERROR), "K-Meleon", MB_ICONEXCLAMATION | MB_OK);
	return FALSE;
}


static BOOL get_node_text(miniCOMPtr<nsIDOMNode>& node,
                          nsAString& value)
{
	nsresult rv;
	miniCOMPtr<nsIDOMNSHTMLElement> element;
	rv = node.QueryInterface(element);
	nsEmbedString t;
	if (NS_SUCCEEDED(rv)) {
		rv = element->GetInnerHTML(t);
	} else {
		rv = node->GetNodeValue(t);
	}
	NS_ENSURE_SUCCESS(rv, FALSE);
	if (! value.IsEmpty()) {
		value += g_delimiter;
	}
	value += t;
	return TRUE;
}


static BOOL set_node_text(miniCOMPtr<nsIDOMNode>& node,
                          nsAString& value)
{
	nsresult rv;
	miniCOMPtr<nsIDOMNSHTMLElement> element;
	rv = node.QueryInterface(element);
	if (NS_SUCCEEDED(rv)) {
		rv = element->SetInnerHTML(value);
	} else {
		rv = node->SetNodeValue(value);
	}
	NS_ENSURE_SUCCESS(rv, FALSE);
	return TRUE;
}


static BOOL process_node(PRUint16 type,
                         miniCOMPtr<nsIDOMXPathResult>& result,
                         nsAString& value,
                         BOOL (*f)(miniCOMPtr<nsIDOMNode>&, nsAString&))
{
	nsresult rv;
	
	switch (type) {
	case nsIDOMXPathResult::UNORDERED_NODE_ITERATOR_TYPE:
	case nsIDOMXPathResult::ORDERED_NODE_ITERATOR_TYPE:
		PRBool state;
		for (;;) {
			rv = result->GetInvalidIteratorState(&state);
			NS_ENSURE_SUCCESS(rv, FALSE);
			if (state) break;
			miniCOMPtr<nsIDOMNode> node;
			rv = result->IterateNext(&node);
			NS_ENSURE_SUCCESS(rv, FALSE);
			if (! node) break;
			if (! (*f)(node, value)) return FALSE;
		}
		break;
	case nsIDOMXPathResult::UNORDERED_NODE_SNAPSHOT_TYPE:
	case nsIDOMXPathResult::ORDERED_NODE_SNAPSHOT_TYPE:
		PRUint32 length;
		rv = result->GetSnapshotLength(&length);
		NS_ENSURE_SUCCESS(rv, FALSE);
		PRUint32 i;
		for (i = 0; i < length; i ++) {
			miniCOMPtr<nsIDOMNode> node;
			rv = result->SnapshotItem(i, &node);
			NS_ENSURE_SUCCESS(rv, FALSE);
			if (! node) break;
			if (! (*f)(node, value)) return FALSE;
		}
		break;
	case nsIDOMXPathResult::ANY_UNORDERED_NODE_TYPE:
	case nsIDOMXPathResult::FIRST_ORDERED_NODE_TYPE:
		{
			miniCOMPtr<nsIDOMNode> node;
			rv = result->GetSingleNodeValue(&node);
			NS_ENSURE_SUCCESS(rv, FALSE);
			NS_ENSURE_TRUE(node, FALSE);
			if (! (*f)(node, value)) return FALSE;
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


char *utils_xpath_get(char *expr)
{
	HWND hwnd;
	nsresult rv;
	
	if (expr == NULL) return NULL;
	if ((hwnd = get_browser_hwnd()) == NULL) return NULL;
	
	miniCOMPtr<nsIDOMXPathResult> result;
	if (! xpath_eval(hwnd, expr, result)) return NULL;
	NS_ENSURE_TRUE(result, NULL);
	
	PRUint16 type;
	rv = result->GetResultType(&type);
	NS_ENSURE_SUCCESS(rv, NULL);
	switch (type) {
	case nsIDOMXPathResult::NUMBER_TYPE:
		{
			double fvalue;
			rv = result->GetNumberValue(&fvalue);
			NS_ENSURE_SUCCESS(rv, NULL);
			PRInt64 lvalue = static_cast<PRInt64>(fvalue);
			char *svalue;
			if (lvalue != fvalue) {
				svalue = PR_smprintf("%f", fvalue);
			} else {
				svalue = PR_smprintf("%lld", lvalue);
			}
			char *r = _strdup(svalue);
			PR_smprintf_free(svalue);
			return r;
		}
	case nsIDOMXPathResult::STRING_TYPE:
		{
			nsEmbedString svalue;
			rv = result->GetStringValue(svalue);
			NS_ENSURE_SUCCESS(rv, NULL);
			return miniMBString(svalue.get()).steal();
		}
	case nsIDOMXPathResult::BOOLEAN_TYPE:
		{
			PRBool bvalue;
			rv = result->GetBooleanValue(&bvalue);
			NS_ENSURE_SUCCESS(rv, NULL);
			return _strdup(bvalue ? "true" : "false");
		}
	}
	nsEmbedString value;
	if (! process_node(type, result, value, get_node_text)) return NULL;
	return miniMBString(value.get()).steal();
}


void utils_xpath_set(char *expr, char *value)
{
	HWND hwnd;
	nsresult rv;
	
	if (expr == NULL || value == NULL) return;
	if ((hwnd = get_browser_hwnd()) == NULL) return;
	
	miniCOMPtr<nsIDOMXPathResult> result;
	if (! xpath_eval(hwnd, expr, result)) return;
	NS_ENSURE_TRUE(result,);
	
	PRUint16 type;
	rv = result->GetResultType(&type);
	NS_ENSURE_SUCCESS(rv,);
	nsEmbedString rep;
	rep = miniWString(value);
	process_node(type, result, rep, set_node_text);
}


void utils_xpath_set_delimiter(char *delimiter, char *dummy)
{
	miniWString s(delimiter);
	if (lstrlenW(s) + 1 > MAX_DELIMITER_LENGTH) return;
	lstrcpyW(g_delimiter, s);
}
